package jalview.ext.ensembl;

import jalview.datamodel.AlignmentI;
import jalview.datamodel.Sequence;
import jalview.datamodel.SequenceFeature;
import jalview.datamodel.SequenceI;
import jalview.io.gff.SequenceOntologyFactory;
import jalview.io.gff.SequenceOntologyI;
import jalview.util.MapList;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import com.stevesoft.pat.Regex;

/**
 * A class that fetches genomic sequence and all transcripts for an Ensembl gene
 * 
 * @author gmcarstairs
 */
public class EnsemblGene extends EnsemblSeqProxy
{
  private static final String GENE_PREFIX = "gene:";

  // TODO modify to accept other species e.g. ENSMUSGnnn
  private static final Regex ACCESSION_REGEX = new Regex(
          "(ENSG|ENST)[0-9]{11}$");

  private static final EnsemblFeatureType[] FEATURES_TO_FETCH = {
      EnsemblFeatureType.gene, EnsemblFeatureType.transcript,
      EnsemblFeatureType.exon, EnsemblFeatureType.cds,
      EnsemblFeatureType.variation };

  @Override
  public String getDbName()
  {
    return "ENSEMBL (GENE)";
  }

  @Override
  protected EnsemblFeatureType[] getFeaturesToFetch()
  {
    return FEATURES_TO_FETCH;
  }

  @Override
  protected EnsemblSeqType getSourceEnsemblType()
  {
    return EnsemblSeqType.GENOMIC;
  }

  /**
   * Builds an alignment of all transcripts for the requested gene:
   * <ul>
   * <li>fetches the gene sequence</li>
   * <li>fetches features on the sequence</li>
   * <li>identifies "transcript" features whose Parent is the requested gene</li>
   * <li>fetches the transcript sequence for each transcript</li>
   * <li>makes a mapping from the gene to each transcript</li>
   * <li>copies features from gene to transcript sequences</li>
   * <li>fetches the protein sequence for each transcript, maps and saves it as
   * a cross-reference</li>
   * <li>aligns each transcript against the gene sequence based on the position
   * mappings</li>
   * </ul>
   */
  @Override
  public AlignmentI getSequenceRecords(String query) throws Exception
  {
    List<String> transcriptsWanted = null;

    if (isTranscriptIdentifier(query))
    {
      transcriptsWanted = Arrays.asList(query
              .split(getAccessionSeparator()));
      query = getGeneForTranscript(query);
      if (query == null)
      {
        return null;
      }
    }

    AlignmentI al = super.getSequenceRecords(query);
    if (al.getHeight() > 0)
    {
      getTranscripts(al, query, transcriptsWanted);
    }

    return al;
  }

  /**
   * Gets the parent gene identifier for a given transcript identifier, by
   * retrieving 'transcript' features overlapping the transcript, and finding
   * the Parent property of the feature whose id is the given identifier.
   * 
   * @param query
   * @return
   */
  protected String getGeneForTranscript(String transcriptId)
  {
    String geneId = null;

    /*
     * reduce multiple transcripts (e.g. from Uniprot x-ref) to the first
     * one only as representative (they should all have the same gene)
     */
    transcriptId = transcriptId.split(getAccessionSeparator())[0];

    try
    {
      EnsemblFeatureType[] geneFeature = new EnsemblFeatureType[] { EnsemblFeatureType.transcript };
      AlignmentI al = new EnsemblFeatures().getSequenceRecords(
              transcriptId, geneFeature);
      if (al != null && al.getHeight() > 0)
      {
        SequenceFeature[] sfs = al.getSequenceAt(0).getSequenceFeatures();
        if (sfs != null)
        {
          for (SequenceFeature sf : sfs)
          {
            if (transcriptId.equals(getTranscriptId(sf)))
            {
              String parent = (String) sf.getValue(PARENT);
              if (parent != null && parent.startsWith(GENE_PREFIX))
              {
                geneId = parent.substring(5);
              }
              break;
            }
          }
        }
      }
      return geneId;
    } catch (IOException e)
    {
      System.err.println("Error retrieving gene id for " + transcriptId
              + ": " + e.getMessage());
      return null;
    }
  }

  /**
   * Constructs all transcripts for the gene, as identified by "transcript"
   * features whose Parent is the requested gene. The coding transcript
   * sequences (i.e. with introns omitted) are added to the alignment.
   * 
   * @param al
   * @param accId
   * @param transcriptsWanted
   *          optional list of transcript ids to filter by
   * @throws Exception
   */
  protected void getTranscripts(AlignmentI al, String accId,
          List<String> transcriptsWanted)
          throws Exception
  {
    SequenceI gene = al.getSequenceAt(0);
    List<SequenceFeature> transcriptFeatures = getTranscriptFeatures(accId,
            gene, transcriptsWanted);

    for (SequenceFeature transcriptFeature : transcriptFeatures)
    {
      makeTranscript(transcriptFeature, al, gene);
    }
  }

  /**
   * Constructs a spliced transcript sequence by finding 'exon' features for the
   * given id (or failing that 'CDS'). Copies features on to the new sequence.
   * 'Aligns' the new sequence against the gene sequence by padding with gaps,
   * and adds it to the alignment.
   * 
   * @param transcriptFeature
   * @param al
   *          the alignment to which to add the new sequence
   * @param gene
   *          the parent gene sequence, with features
   * @return
   */
  SequenceI makeTranscript(SequenceFeature transcriptFeature,
          AlignmentI al, SequenceI gene)
  {
    String accId = getTranscriptId(transcriptFeature);
    if (accId == null)
    {
      return null;
    }

    /*
     * NB we are mapping from gene sequence (not genome), so do not
     * need to check for reverse strand (gene and transcript sequences 
     * are in forward sense)
     */

    /*
     * make a gene-length sequence filled with gaps
     * we will fill in the bases for transcript regions
     */
    char[] seqChars = new char[gene.getLength()];
    Arrays.fill(seqChars, al.getGapCharacter());

    /*
     * look for exon features of the transcript, failing that for CDS
     * (for example ENSG00000124610 has 1 CDS but no exon features)
     */
    String parentId = "transcript:" + accId;
    List<SequenceFeature> splices = findFeatures(gene,
            SequenceOntologyI.EXON, parentId);
    if (splices.isEmpty())
    {
      splices = findFeatures(gene, SequenceOntologyI.CDS, parentId);
    }

    int transcriptLength = 0;
    final char[] geneChars = gene.getSequence();
    int offset = gene.getStart(); // to convert to 0-based positions
    List<int[]> mappedFrom = new ArrayList<int[]>();

    for (SequenceFeature sf : splices)
    {
      int start = sf.getBegin() - offset;
      int end = sf.getEnd() - offset;
      int spliceLength = end - start + 1;
      System.arraycopy(geneChars, start, seqChars, start, spliceLength);
      transcriptLength += spliceLength;
      mappedFrom.add(new int[] { sf.getBegin(), sf.getEnd() });
    }

    Sequence transcript = new Sequence(accId, seqChars, 1, transcriptLength);
    String geneName = (String) transcriptFeature.getValue(NAME);
    if (geneName != null)
    {
      transcript.setDescription(geneName);
    }
    transcript.createDatasetSequence();

    al.addSequence(transcript);

    /*
     * transfer features to the new sequence; we use EnsemblCdna to do this,
     * to filter out unwanted features types (see method retainFeature)
     */
    List<int[]> mapTo = new ArrayList<int[]>();
    mapTo.add(new int[] { 1, transcriptLength });
    MapList mapping = new MapList(mappedFrom, mapTo, 1, 1);
    new EnsemblCdna().transferFeatures(gene.getSequenceFeatures(),
            transcript.getDatasetSequence(), mapping, parentId);

    /*
     * and finally fetch the protein product and save as a cross-reference
     */
    new EnsemblCdna().addProteinProduct(transcript);

    return transcript;
  }

  /**
   * Returns the 'transcript_id' property of the sequence feature (or null)
   * 
   * @param feature
   * @return
   */
  protected String getTranscriptId(SequenceFeature feature)
  {
    return (String) feature.getValue("transcript_id");
  }

  /**
   * Returns a list of the transcript features on the sequence whose Parent is
   * the gene for the accession id.
   * 
   * @param accId
   * @param geneSequence
   * @param transcriptsWanted
   *          optional list of ids to filter on
   * @return
   */
  protected List<SequenceFeature> getTranscriptFeatures(String accId,
          SequenceI geneSequence, List<String> transcriptsWanted)
  {
    List<SequenceFeature> transcriptFeatures = new ArrayList<SequenceFeature>();

    String parentIdentifier = GENE_PREFIX + accId;
    SequenceFeature[] sfs = geneSequence.getSequenceFeatures();

    if (sfs != null)
    {
      for (SequenceFeature sf : sfs)
      {
        if (isTranscript(sf.getType()))
        {
          if (transcriptsWanted != null)
          {
            String transcriptId = (String) sf.getValue("transcript_id");
            if (!transcriptsWanted.contains(transcriptId))
            {
              // continue;
            }
          }
          String parent = (String) sf.getValue(PARENT);
          if (parentIdentifier.equals(parent))
          {
            transcriptFeatures.add(sf);
          }
        }
      }
    }

    return transcriptFeatures;
  }

  @Override
  public String getDescription()
  {
    return "Fetches all transcripts and variant features for a gene or transcript";
  }

  /**
   * Default test query is a gene id (can also enter a transcript id)
   */
  @Override
  public String getTestQuery()
  {
    return "ENSG00000157764"; // BRAF, 5 transcripts, reverse strand
    // ENSG00000090266 // NDUFB2, 15 transcripts, forward strand
    // ENSG00000101812 // H2BFM histone, 3 transcripts, forward strand
    // ENSG00000123569 // H2BFWT histone, 2 transcripts, reverse strand
  }

  /**
   * Answers true for a feature of type 'gene' (or a sub-type of gene in the
   * Sequence Ontology), whose ID is the accession we are retrieving
   */
  @Override
  protected boolean identifiesSequence(SequenceFeature sf, String accId)
  {
    if (SequenceOntologyFactory.getInstance().isA(sf.getType(),
            SequenceOntologyI.GENE))
    {
      String id = (String) sf.getValue(ID);
      if ((GENE_PREFIX + accId).equals(id))
      {
        return true;
      }
    }
    return false;
  }

  /**
   * Answers true unless feature type is 'gene', or 'transcript' with a parent
   * which is a different gene. We need the gene features to identify the range,
   * but it is redundant information on the gene sequence. Checking the parent
   * allows us to drop transcript features which belong to different
   * (overlapping) genes.
   */
  @Override
  protected boolean retainFeature(SequenceFeature sf, String accessionId)
  {
    if (SequenceOntologyFactory.getInstance().isA(sf.getType(),
            SequenceOntologyI.GENE))
    {
      return false;
    }

    if (isTranscript(sf.getType()))
    {
      String parent = (String) sf.getValue(PARENT);
      if (!(GENE_PREFIX + accessionId).equals(parent))
      {
        return false;
      }
    }
    return true;
  }

  /**
   * Answers false. This allows an optimisation - a single 'gene' feature is all
   * that is needed to identify the positions of the gene on the genomic
   * sequence.
   */
  @Override
  protected boolean isSpliceable()
  {
    return false;
  }

  @Override
  protected List<String> getCrossReferenceDatabases()
  {
    // found these for ENSG00000157764 on 30/01/2016:
    // return new String[] {"Vega_gene", "OTTG", "ENS_LRG_gene", "ArrayExpress",
    // "EntrezGene", "HGNC", "MIM_GENE", "MIM_MORBID", "WikiGene"};
    return super.getCrossReferenceDatabases();
  }

  /**
   * Override to do nothing as Ensembl doesn't return a protein sequence for a
   * gene identifier
   */
  @Override
  protected void addProteinProduct(SequenceI querySeq)
  {
  }

  @Override
  public Regex getAccessionValidator()
  {
    return ACCESSION_REGEX;
  }

}
